signature DEBUGUTILS = sig

(* Debugging utilities *)

(* A "debugging wrapper" for (optionally) printing the arguments and
   results to functions, maintaining proper nesting and indentation. *)
val wrap: 
      (bool ref)         (* print debugging info only when set to true *)
   -> string             (* name of debugged function *)
   -> (unit -> 'a)       (* thunked call to underlying function *)
   -> (unit -> unit)     (* called before underlying fn to show args *)
   -> ('a -> unit)       (* used to show result returned by underlying fn *)
   -> 'a

(* Like wrap, but takes a list of flags and proceeds if ANY of them is 
   set to true *)
val wrapl: 
      (bool ref list) 
   -> string 
   -> (unit -> 'a) 
   -> (unit -> unit) 
   -> ('a -> unit) 
   -> 'a

(* Internal reset routine *)
val reset_debugutils: unit -> unit

end
