% First we introduce a new Object type constructor, which
% makes explicit the fact that the method vector of an object
% is still abstracted on a final representation type and a 
% vector of "self" methods: 
Object = Fun(M:*->*) 
            Some(R) 
              {state: R, 
               methods: All(SubR)
                        (SubR->R)
                     -> (SubR->R->SubR)
                     -> M(SubR)
                     -> M(SubR)};

% Now if we take the standard pair of object interfaces -- points
% and colored points with a slide method...
PointM = Fun(R) {setX: R->Int->R, getX: R->Int, slide: R->Int->R};
CPointM = Fun(R) {setX: R->Int->R, getX: R->Int, slide: R->Int->R, 
                   setC: R->Color->R, getC: R->Color};

% we can implement the operations of sending various messages to
% objects as follows.  (Basically, we have just moved the knot-typing
% code from the "new" combinator into the message sending functions.)
Point'slide = 
  fun(M < PointM)
  fun(p: Object(M))
    fun(i:Int)
      open p as <R,r> in
        let m = rec [M(R)] 
                  fun(mr:M(R))
                    r.methods 
                      [R]
                      (fun(r:R) r)
                      (fun(r:R) fun(r':R) r')
                      mr
        in pack 
          {state = m.slide r.state i,
           methods = r.methods}
        as Object(M) hiding R
        end end
      end;

Point'getX = 
  fun(M < PointM)
    fun(p: Object(M))
      open p as <R,r> in
        let m = rec [M(R)] 
                  fun(mr:M(R))
                    r.methods 
                      [R]
                      (fun(r:R) r)
                      (fun(r:R) fun(r':R) r')
                      mr
        in m.getX r.state
        end
      end;

% A typical instance of the Point class is:
p = pack
      {state = 5,
       methods = fun(SubR)
                 fun(get: SubR->Int)
                 fun(put: SubR->Int->SubR)
                 fun(m: PointM(SubR))
                    {setX = fun(r:SubR) fun(i:Int) put r i,
                     getX = fun(r:SubR) get r,
                     slide = fun(r:SubR) fun(i:Int) 
                              m.setX r (plus (m.getX r) i)}
       }
    as Object(PointM) hiding Int
    end;

% Programs that manipulate points by sending them messages look
% exactly the same as before:
Point'getX [PointM] (Point'slide [PointM] p 3);

% Now we can build a colored point c that delegates the Point operations 
% slide and getX to p, while, for the sake of the example, providing its 
% own setX that changes the color to red.
c = open p as <R,r> in
    let MyR = {mystate:Color, delegated_state: R} in
    pack
      {state = {mystate = blue, delegated_state = r.state},
       methods = fun(SubR)
                 fun(get: SubR->MyR)
                 fun(put: SubR->MyR->SubR)
                 fun(m: CPointM(SubR))
                    {setC = fun(r:SubR) fun(c:Color) 
                              put r {mystate=c, 
                                     delegated_state=(get r).delegated_state},
                     getC = fun(r:SubR) (get r).mystate,
                     setX = fun(r:SubR) 
                              m.setX 
                               (put r 
                                  {mystate=red, 
                                   delegated_state=(get r).delegated_state}),
                     getX = m.getX,
                     slide = m.slide}
       }
    as Object(CPointM) hiding MyR
    end end end;


