CPointM = Fun(Rep) {|getX: Rep->Int, setX: Rep->Int->Rep, 
                      getC: Rep->Color, setC: Rep->Color->Rep|};

Package =
  Fun(SelfM: *->*)  % Method interface
  Fun(SelfF: *->*)  % Binary function interface
    Some(InternalM<SelfM)             % Hidden internal interface
     {|init: Object(InternalM),       % Initial object
       fns: SelfF(Object InternalM)|};% Binary functions

ClassResult =
  Fun(SelfH: *->*) % Method interface including rep method 
  Fun(SelfF: *->*) % Binary function interface
  Fun(FinalR: *)   % Final representation
    {|methods: SelfH(FinalR), fns: SelfF(Object SelfH)|};

Class = 
  Fun(SelfM: *->*) % object interface type
  Fun(SelfH: *->*) % object interface extended with rep method
  Fun(SelfF: *->*) % binary function interface
  Fun(SelfR: *)    % internal representation type
    All(FinalR)                     % final representation
    (FinalR->SelfR) ->              % extractor
    (FinalR->SelfR->FinalR) ->      % overwriter
    (SelfM(FinalR)) ->              % self methods
    ClassResult SelfH SelfF FinalR; % Class result

instantiate =
  fun(SelfM < Top(*->*)) % object interface type
  fun(SelfH < SelfM)     % object interface extended with rep method
  fun(SelfF < Top(*->*)) % binary methods
  fun(SelfR < Top(*))    % internal representation type
  fun(s : SelfR)         % initial state
  fun(selfClass : Class SelfM SelfH SelfF SelfR)
    let self =
      rec (ClassResult SelfH SelfF SelfR)
        fun(self:ClassResult SelfH SelfF SelfR)
          selfClass SelfR (fun(s:SelfR) s)
          (fun(s:SelfR) fun(s':SelfR) s') self.methods
    in <SelfH,
        {init =
           <SelfR,
            {state = s, methods = self.methods}
           >: Object(SelfH),
         fns = self.fns}
       >: Package SelfM SelfF
    end;

%%% !besilent;
getRep = 
  fun(R < Top(*))
  fun(ob: Object(Fun(Rep) {|rep: Rep -> R|}))
    open ob as <Rep,r> in
      r.methods.rep r.state
    end;

Point =
  Objecttype(Rep) with
    getX: Int,
    setX: Int->Rep
  end;
%%% !beloud;

PointR = {|x:Int|};
PointM = Fun(Rep) {|setX:Rep->Int->Rep, getX:Rep->Int|};
PointH = Fun(Rep) {|setX:Rep->Int->Rep, getX:Rep->Int, rep:Rep->PointR|};
PointF = Fun(Obj) {|eq:Obj->Obj->Bool|};

pointClass =
 (fun(FinalR<Top(*))
  fun(get: FinalR->PointR)
  fun(put: FinalR->PointR->FinalR)
  fun(self: PointM(FinalR))
    {methods = 
       {setX = fun(s:FinalR) fun(i:Int) put s {x=i},
        getX = fun(s:FinalR) (get s).x,
        rep  = fun(s:FinalR) get s},
     fns = 
       {eq = fun(p1: Object(PointH)) fun(p2: Object(PointH))
          eqInt (getRep PointR p1).x (getRep PointR p2).x}
    }):< Class PointM PointH PointF PointR >;

pointPackage = 
  instantiate PointM PointH PointF PointR
  {x=5} pointClass;

open pointPackage as <PointM,point> in
  let p = point.init in
    point.fns.eq p (Point'setX PointM p 1)
  end
end;

Increment =
 Fun(SuperM: *->*) Fun(SuperH: *->*) Fun(SuperF: *->*)
 Fun(NewM:   *->*) Fun(NewH:   *->*) Fun(NewF:   *->*) Fun(NewR: *)
  All(FinalR)                     % Final representation
    (FinalR->NewR) ->             % extractor
    (FinalR->NewR->FinalR) ->     % overwriter
    (SuperM(FinalR)) ->           % Superclass methods
    (SuperF(Object SuperH)) ->    % Superclass functions
    (NewM(FinalR)) ->             % Self methods
    ClassResult NewH NewF FinalR; % Class result

extend = 
  fun(SuperM < Top(*->*)) % Superclass interface
  fun(SuperH < SuperM)    % Superclass interface extended with rep method
  fun(SuperF < Top(*->*)) % Superclass binary functions
  fun(SuperR < Top(*))    % Superclass internal representation type
  fun(NewM < SuperM)      % New class interface
  fun(NewH < NewM)        % New class interface extended with rep method
  fun(NewF < SuperF)      % New class binary functions
  fun(NewR < Top(*))      % New class internal representation type
  fun(superClass: Class SuperM SuperH SuperF SuperR) % The superclass
  fun(inc: Increment SuperM SuperH     % The increment function
           SuperF NewM NewH NewF NewR)
  fun(get: NewR->SuperR)       % new->super extractor
  fun(put: NewR->SuperR->NewR) % new<-super overwriter
    % Build the extended class...
   (fun(FinalR)
    fun(g: FinalR->NewR)
    fun(p: FinalR->NewR->FinalR)
    fun(newM: NewM(FinalR))
      let super =
        superClass FinalR
        (fun(s:FinalR) get(g s))
        (fun(s:FinalR) fun(s':SuperR) p s (put (g s) s'))
        newM
      in inc FinalR g p super.methods super.fns newM
      end):< Class NewM NewH NewF NewR >;

CPointR = {|x:Int,c:Color|};
CPointM = Fun(Rep) {|setX:Rep->Int->Rep, getX:Rep->Int,
                      setC:Rep->Color->Rep, getC:Rep->Color|};
CPointH = Fun(Rep) {|setX:Rep->Int->Rep, getX:Rep->Int, 
                      setC:Rep->Color->Rep, getC:Rep->Color,
                      rep:Rep->CPointR|};
CPointF = Fun(Obj) {|eq:Obj->Obj->Bool|};

% A cpoint class that inherits the equality method of points

cpointClass = 
  extend
  PointM PointH PointF PointR
  CPointM CPointH CPointF CPointR
  pointClass
  (fun(FinalR:*)
   fun(get: FinalR->CPointR)
   fun(put: FinalR->CPointR->FinalR)
   fun(super: PointM(FinalR))
   fun(superf: PointF(Object PointH))
   fun(self: CPointM(FinalR))
     {methods = 
        {setX = super.setX,
         getX = super.getX,
         setC = fun(s:FinalR) fun(newc:Color) put s {x=(get s).x, c=newc},
         getC = fun(s:FinalR) (get s).c,
         rep  = fun(s:FinalR) get s},
      fns = 
        {eq = superf.eq}})
  (fun(s:CPointR) {x=s.x})
  (fun(s:CPointR) fun(new:{|x:Int|}) {x=new.x, c=s.c});

cpointPackage = 
  instantiate CPointM CPointH CPointF CPointR
  {x=5,c=red} cpointClass;

open cpointPackage as <CPointM,cpoint> in
  let p = cpoint.init in
    cpoint.fns.eq p (Point'setX CPointM p 1)
  end
end;

% Now the better version, with its own equality method

cpointClass = 
  extend
  PointM PointH PointF PointR
  CPointM CPointH CPointF CPointR
  pointClass
  (fun(FinalR<Top(*))
   fun(get: FinalR->CPointR)
   fun(put: FinalR->CPointR->FinalR)
   fun(super: PointM(FinalR))
   fun(superf: PointF(Object PointH))
   fun(self: CPointM(FinalR))
     {methods = 
        {setX = super.setX,
         getX = super.getX,
         setC = fun(s:FinalR) fun(newc:Color) put s {x=(get s).x, c=newc},
         getC = fun(s:FinalR) (get s).c,
         rep  = fun(s:FinalR) get s},
      fns = 
        {eq = fun(p1: Object(CPointH)) fun(p2: Object(CPointH))
               and
                (eqInt (getRep CPointR p1).x (getRep CPointR p2).x)
                (eqColor (getRep CPointR p1).c (getRep CPointR p2).c)
        }})
  (fun(s:CPointR) {x=s.x})
  (fun(s:CPointR) fun(new:{|x:Int|}) {x=new.x, c=s.c});

cpointPackage = 
  instantiate CPointM CPointH CPointF CPointR
  {x=5,c=red} cpointClass;

open cpointPackage as <CPointM,cpoint> in
  let p = cpoint.init in
    cpoint.fns.eq p (Point'setX CPointM p 1)
  end
end;

