structure Tokens = Tokens
structure Interface = Interface
open Interface

type pos = Interface.pos
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

val eof = fn () => Tokens.T_EOF(!line,!line)

val str_begin = ref(!line);
val str_const = ref([]:string list);

%%
%header (functor MYLexFun(structure Tokens: MY_TOKENS structure Interface: INTERFACE) : LEXER);

%s COMMENT STRING;
vidinit=[a-z0-9#^-];
tidinit=[A-Z];
vidchars=[A-Za-z0-9_!#^$'];
tidchars=[A-Za-z0-9_!#^$];
digit=[0-9];
signordigit=[0-9~];
vid={vidinit}{vidchars}*;
tid={tidinit}{tidchars}*;
compoundvid={tid}"'"{vid};
int={signordigit}{digit}*;
ws=[\t\ ]*;   
%%
<INITIAL>{ws}	=> (lex());
<INITIAL>\n	=> (next_line(); lex());
<INITIAL>"%"	=> (YYBEGIN COMMENT; lex());
<INITIAL>":"	=> (Tokens.T_COLON(!line,!line));
<INITIAL>":<"	=> (Tokens.T_COLONLANGLE(!line,!line));
<INITIAL>"*"	=> (Tokens.T_STAR(!line,!line));
<INITIAL>"@"	=> (Tokens.T_AT(!line,!line));
<INITIAL>";"    => (Tokens.T_EOF(!line,!line));
<INITIAL>"."    => (Tokens.T_DOT(!line,!line));
<INITIAL>"!"    => (Tokens.T_BANG(!line,!line));
<INITIAL>"~"    => (Tokens.T_TILDE(!line,!line));
<INITIAL>","    => (Tokens.T_COMMA(!line,!line));
<INITIAL>"="    => (Tokens.T_EQ(!line,!line));
<INITIAL>"=="   => (Tokens.T_DOUBLEEQ(!line,!line));
<INITIAL>"<="   => (Tokens.T_LEQ(!line,!line));
<INITIAL>"<"    => (Tokens.T_LEQ(!line,!line));
<INITIAL>">"	=> (Tokens.T_RANGLE(!line,!line));
<INITIAL>"("	=> (Tokens.T_LPAREN(!line,!line));
<INITIAL>")"	=> (Tokens.T_RPAREN(!line,!line));
<INITIAL>"["	=> (Tokens.T_LBRACK(!line,!line));
<INITIAL>"]"	=> (Tokens.T_RBRACK(!line,!line));
<INITIAL>"<<"	=> (Tokens.T_DOUBLELANGLE(!line,!line));
<INITIAL>">>"	=> (Tokens.T_DOUBLERANGLE(!line,!line));
<INITIAL>"{|"	=> (Tokens.T_DOUBLELCURLY(!line,!line));
<INITIAL>"|}"	=> (Tokens.T_DOUBLERCURLY(!line,!line));
<INITIAL>"{"	=> (Tokens.T_LCURLY(!line,!line));
<INITIAL>"}"	=> (Tokens.T_RCURLY(!line,!line));
<INITIAL>"->"	=> (Tokens.T_ARROW(!line,!line));
<INITIAL>"\\\\"	=> (Tokens.T_BIGLAMBDA(!line,!line));
<INITIAL>"\\"	=> (Tokens.T_LAMBDA(!line,!line));
<INITIAL>"\""   => (str_begin:=(!line); str_const:=[]; YYBEGIN STRING; lex());
<INITIAL>{int}  => (Tokens.T_INT_CONST (yytext,!line,!line));
<INITIAL>{compoundvid}  => (Tokens.T_VID (yytext,!line,!line));
<INITIAL>{vid}  => (case yytext of
		       "as" => Tokens.T_AS(!line,!line)
		     | "begin" => Tokens.T_BEGIN (!line,!line)
		     | "check" => Tokens.T_CHECK(!line,!line)
		     | "debug" => Tokens.T_DEBUG(!line,!line)
		     | "end" => Tokens.T_END (!line,!line)
		     | "fun" => Tokens.T_LAMBDA(!line,!line)
		     | "hiding" => Tokens.T_HIDING(!line,!line)
		     | "in" => Tokens.T_IN(!line,!line)
		     | "if" => Tokens.T_IF(!line,!line)
		     | "then" => Tokens.T_THEN(!line,!line)
		     | "else" => Tokens.T_ELSE(!line,!line)
		     | "let" => Tokens.T_LET (!line,!line)
		     | "open" => Tokens.T_OPEN(!line,!line)
		     | "pack" => Tokens.T_PACK(!line,!line)
		     | "reset" => Tokens.T_RESET(!line,!line)
		     | "set" => Tokens.T_SET(!line,!line)
		     | "type" => Tokens.T_TYPE(!line,!line)
		     | "with" => Tokens.T_WITH(!line,!line)
                     | "use" => Tokens.T_USE(!line,!line)
                     | "class" => Tokens.T_CLASS(!line,!line)
                     | "vars" => Tokens.T_VARS(!line,!line)
                     | "inherit" => Tokens.T_INHERIT(!line,!line)
                     | "from" => Tokens.T_FROM(!line,!line)
                     | "virtual" => Tokens.T_VIRTUAL(!line,!line)
                     | "printer" => Tokens.T_PRINTER(!line,!line)
                     | _ => Tokens.T_VID (yytext,!line,!line));
<INITIAL>{tid}  => (case yytext of
		       "All" => Tokens.T_ALL(!line,!line)
		     | "Fun" => Tokens.T_BIGLAMBDA(!line,!line)
		     | "Some" => Tokens.T_SOME(!line,!line)
		     | "Top" => Tokens.T_TOP(!line,!line)
                     | "Objecttype" => Tokens.T_OBJECTTYPE(!line,!line)
                     | "ObjectType" => Tokens.T_OBJECTTYPE(!line,!line)
                     | _ => Tokens.T_TID (yytext,!line,!line));
<INITIAL>.	=> (error ("ignoring illegal character" ^ yytext,
			   !line,!line); lex());

<COMMENT>\n     => (next_line(); YYBEGIN INITIAL; lex());
<COMMENT>.	=> (lex());

<STRING>\n      => (next_line(); lex());
<STRING>\\\"	=> (str_const:=("\""::(!str_const)); lex());
<STRING>"\""    => (YYBEGIN INITIAL;
		    Tokens.T_STR_CONST(implode(rev(!str_const)),
		    		       !str_begin,!line));
<STRING>.	=> (str_const:=(yytext::(!str_const)); lex());

